/*============================================================================
: LiFe
:     
: - (  sm_)
: ..
  : 28.02.2003
==============================================================================*/
#include "_main.h"

#include <assert.h>
#include <stdlib.h>
#include <stdio.h>

#include "sm_main.h"

/* ==========================   =======================*/
/* ==================================  ===============================*/
void 	_delColumn(int j,int*n_,int m,int*fV,float*T);
void	_delRow(int i,int*n_,int*m_,int*fV,float*T);
void	_replaceBasis(int i_,int j_,int n,int m,int*bV,int*fV,float*T );
void 	_RoundToZero(int m,int k,float*T);
/* ============================   =======================*/
/* --------------------------------- sm_getX -------------------------------- */
/*   -   x.    x[0] 
	  - */
void sm_getX( /* : */
	int n, 				/*  . - */
	int m, 				/* .  - */
	int *bV,		/*  - - (   1) */
	float *T,	/* - */
	float *x	/*  . x[0]    - */
){
#define a(i,j) T[(i)*SM_MAX_DIM+j]
   int i,j;
   for(j=0; j<=n; j++)
   	x[j]=0; /* -   . */
   for(i=1; i<=m; i++){
   	x[bV[i]]=a(i,0);
  	x[0]=-a(0,0);
	};
#undef a
};

/* ------------------------------ sm_canonView ------------------------------ */
/*  -       
    . */
int /* :
			 1 -    ;
          0 -    */
sm_canonView( /* : */
	int n, 			/*  . -   */
	int m, 			/* .    */
   float *A,	/*    
                  A[0,0]=-Q_0;
                  A[0,1..n] -    (k=n-m);
                  A[1..m,0] -  b_i >= 0;
                  A[1..m,1..n] -   */

	int *n_, 			/*  . -  - */
	int *m_, 			/* .  -  - */
	int *bV,		/*  - - (   1) */
	int *fV,		/*  - - (   1) */
	float *T	/* -,   :
                  T_[0,0]=-Q_0;
                  T_[0,1..k] -    (k=n-m);
                  T_[1..m,0] -  b_i;
                  T_[1..m,1..k] -   */
){
#define a(i,j)  A[(i)*SM_MAX_DIM+j]
#define t(i,j) T[(i)*SM_MAX_DIM+j]
#define _n *n_
#define _m *m_
#define _k (_n-_m)
   int i,j;
	int done,OK;
   float Q[SM_MAX_DIM];

   _n=n+m;
   assert(_n<SM_MAX_DIM);
   _m=m;

   for(i=1;i<=_m;i++) /*        ( ..!) */
   	for(j=0;j<=n;j++)
      	t(i,j)=a(i,j);

   for(i=1;i<=_m;i++) /*     */
   	bV[i]=n+i;

  	for(j=1;j<=_k;j++) /*     */
   	fV[j]=j;

  	for(j=0;j<=_k;j++){ /*     */
      t(0,j)=0;
		for(i=1;i<=_m;i++)
			t(0,j)-=a(i,j);
   };

	/* ---------    ----------> */
   do{
	   done=sm_symplex(_n,_m,bV,fV,T);
		if (done==-1){
			return 0;
   	};

		for(j=1;j<=_k;j++)
      	if(fV[j]>n) /*      */
         	_delColumn(j,n_,_m,fV,T);
   }while(!done);

	//  ,     _DELTAAIJ
	_RoundToZero(_m,_k,T);/**/

   if(t(0,0)<0 && !_EQZERO(t(0,0),_DELTACANONVIEW)) //   ! --------------------~~~~~~~~!!!!!!!!!!
    	return 0;

	for(i=1;i<=_m;i++) /*      */
   	if(bV[i]>n){ /* i-   ! */

      	OK=1;
         for(j=1;j<=_k;j++) /*    i-  ? */
         	if(t(i,j)!=0){
            	OK=0;
               break;
            };

         if(OK){	/*    i-    */
         	_delRow(i,n_,m_,bV,T); assert(t(i,0)==0);
         }else{	/* j-   i-     */
         	_replaceBasis(i,j,_n,_m,bV,fV,T);
         	_delColumn(j,n_,_m,fV,T);
         };
   	};

   /*  ,     _DELTAAIJ */
	_RoundToZero(_m,_k,T);

   /*          */
   for(j=0;j<=n;j++) /*    */
   	Q[j]=a(0,j);

   /*  ,     _DELTAAIJ */
	_RoundToZero(_m,_k,T);


   for(i=1;i<=_m;i++) /*        */
   	if(Q[bV[i]]!=0){ /*  i-   */
      	for(j=0;j<=_k;j++)
				Q[fV[j]]-=t(i,j)*Q[bV[i]];
         Q[bV[i]]=0;
		};

   /*  Q  T */
  	for(j=0;j<=_k;j++)
   	t(0,j)=Q[fV[j]];

	return 1;
#undef a
#undef t
#undef _n
#undef _k
};

/*------------------------------- sm_symplex -----------------------------------*/
/*    - */
int /* :
			 1 -  ;
          0 -   ;
         -1 -    */
sm_symplex( /* : */
int n, 				/*  . - */
int m, 				/* .  - */
int *bV,		/*  - - (   1) */
int *fV,		/*  - - (   1) */
float *T) /* - (   ),
						  :
                  T[0,0]=-Q_0;
                  T[0,1..k] -    (k=n-m);
                  T[1..m,0] -  b_i;
                  T[1..m,1..k] -   */
#define t(i,j) T[(i)*SM_MAX_DIM+j]
#define b(i) T[(i)*SM_MAX_DIM]
#define p(j) T[j]
#define Q T[0]
{
	int _j;	/*    */
	int _i;	/*    */
   int k=n-m;
   int i, j;
   float x,y;


   /*    */
   _j=0;
   x=0;
   for (j=1; j<=k; j++)
   	if(p(j)<x && !_EQZERO(p(j),_DELTACOL)){
      	x=p(j);
         _j=j;
      };
   if(_j==0)
   	return 1;

   /*    */
   _i=0;
   x=-1;
   for (i=1; i<=m; i++)
   	if(t(i,_j)>0 && !_EQZERO(t(i,_j),_DELTAROW)){
      	y=b(i)/t(i,_j);
			if((x<0)|(y<x)){
         	x=y;
            _i=i;
         };
      };

   if(_i==0)
   	return -1;

/*   */
/*(t)*/
   i=bV[_i];
   bV[_i]=fV[_j];
	fV[_j]=i;

/*()*/
	t(_i,_j)=1/t(_i,_j);

/*()*/
	for(j=0; j<=k; j++)
   	if(j!=_j)
	   	t(_i,j)=t(_i,j)*t(_i,_j);

/*()*/
	for(i=0; i<=m; i++)
   	if(i!=_i){
			for(j=0; j<=k; j++)
	   		if(j!=_j)
         		t(i,j)=t(i,j)-t(_i,j)*t(i,_j);
	   	t(i,_j)=-t(i,_j)*t(_i,_j);
	   };

   /*  ,     _EPS! */
	_RoundToZero(m,k,T);

   return 0;
#undef t
#undef b
#undef p
#undef Q
};

/* =============================   ==========================*/

/*------------------------------ _delColumn ----------------------------------*/
/*  j-   - */
static void _delColumn(int j,int *n_,int m,int *fV,float *T){
#define t(i,j) T[(i)*SM_MAX_DIM+j]
#define n *n_
#define k (n-m)
   int i,l;

   for(i=0;i<=m;i++)
   	for(l=j;l<k;l++)
      	t(i,l)=t(i,l+1);

  	for(l=j;l<k;l++)
     	fV[l]=fV[l+1];

   n=n-1;
#undef t
#undef n
#undef k
};

/*------------------------------ _delRow ----------------------------------*/
/*  i-   - */
static void _delRow(int i,int*n_,int*m_,int*bV,float*T){
#define t(i,j) T[(i)*SM_MAX_DIM+j]
#define n *n_
#define m *m_
#define k (n-m)
   int j,l;

   for(j=0;j<=k;j++)
   	for(l=i;l<m;l++)
      	t(l,j)=t(l+1,j);

  	for(l=i;l<m;l++)
     	bV[l]=bV[l+1];

   n=n-1;
   m=m-1;
#undef t
#undef n
#undef m
#undef k
};

/*---------------------------- _replaceBasis ---------------------------------*/
/*   */
static void _replaceBasis(int i_,int j_,int n,int m,int*bV,int*fV,float*T){
#define t(i,j) T[(i)*SM_MAX_DIM+j]
#define k (n-m)
	int i, j;

/*(t)*/
   i=bV[i_];
   bV[i_]=fV[j_];
	fV[j_]=i;

/*()*/
	t(i_,j_)=1/t(i_,j_);

/*()*/
	for(j=0; j<=k; j++)
   	if(j!=j_)
	   	t(i_,j)=t(i_,j)*t(i_,j_);

/*()*/
	for(i=0; i<=m; i++)
   	if(i!=i_){
			for(j=0; j<=k; j++)
	   		if(j!=j_)
         		t(i,j)=t(i,j)-t(i_,j)*t(i,j_);
	   	t(i,j_)=-t(i,j_)*t(i_,j_);
	   };

#undef k
#undef t
};

/*    Tmk ,     _DELTAAIJ */
static void _RoundToZero(int m,int k,float*T){
#define t(i,j) T[(i)*SM_MAX_DIM+j]
	int i, j;
	for(i=0; i<=m; i++)
   	for(j=0;j<=k;j++){
      	if(_EQZERO(t(i,j),_DELTAAIJ))
				t(i,j)=0;
      };
#undef t
};

